import 'package:flutter/services.dart';

import '../internal/chat_method_keys.dart';
import '../tools/em_extension.dart';
import 'em_chat_enums.dart';
import 'em_error.dart';

class EMPushConfigs {
  EMPushConfigs._private();

  EMPushStyle? _pushStyle;
  bool? _noDisturb;
  int? _noDisturbStartHour;
  int? _noDisturbEndHour;
  List<String>? _noDisturbGroups = [];

  EMPushStyle? get pushStyle => _pushStyle;
  bool? get noDisturb => _noDisturb;
  int? get noDisturbStartHour => _noDisturbStartHour;
  int? get noDisturbEndHour => _noDisturbEndHour;
  List<String>? get noDisturbGroups => _noDisturbGroups;

  factory EMPushConfigs.fromJson(Map map) {
    return EMPushConfigs._private()
      .._pushStyle =
          map['pushStyle'] == 0 ? EMPushStyle.Simple : EMPushStyle.Summary
      .._noDisturb = map.boolValue('noDisturb')
      .._noDisturbStartHour = map['noDisturbStartHour']
      .._noDisturbEndHour = map['noDisturbEndHour'];
  }

  Map toJson() {
    Map data = Map();
    data['pushStyle'] = _pushStyle == EMPushStyle.Simple;
    data['noDisturb'] = _noDisturb;
    data['noDisturbStartHour'] = _noDisturbStartHour;
    data['noDisturbEndHour'] = _noDisturbEndHour;
    return data;
  }
}

extension EMPushConfigsExtension on EMPushConfigs {
  // channel的命名与pushManager中的channel一致，本质上还是一个channel。
  static const MethodChannel _channel =
      const MethodChannel('com.chat.im/chat_push_manager', JSONMethodCodec());

  /// 设置是否免打扰[isNoDisturb], [startTime], [endTime]
  Future<void> setNoDisturb(
    bool isNoDisturb, [
    int startTime = 0,
    int endTime = 24,
  ]) async {
    if (startTime < 0) startTime = 0;
    if (endTime > 24) endTime = 24;
    Map req = {
      'noDisturb': isNoDisturb,
      'startTime': startTime,
      'endTime': endTime
    };
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.imPushNoDisturb, req);
    try {
      EMError.hasErrorFromResult(result);
      bool success = result.boolValue(ChatMethodKeys.imPushNoDisturb);
      if (success) {
        _noDisturb = isNoDisturb;
        _noDisturbStartHour = startTime;
        _noDisturbEndHour = endTime;
      }
    } on EMError catch (e) {
      throw e;
    }
  }

  /// 设置消息推送显示样式[pushStyle]
  Future<void> setPushStyle(EMPushStyle pushStyle) async {
    Map req = {'pushStyle': pushStyle == EMPushStyle.Simple ? 0 : 1};
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.updateImPushStyle, req);
    try {
      EMError.hasErrorFromResult(result);
    } on EMError catch (e) {
      throw e;
    }
  }

  /// 通过群id[groupId]设置群组是否免打扰[isNoDisturb]
  Future<void> setGroupToDisturb(
    String groupId,
    bool isNoDisturb,
  ) async {
    Map req = {'noDisturb': isNoDisturb, 'group_id': groupId};
    Map result =
        await _channel.invokeMethod(ChatMethodKeys.updateGroupPushService, req);
    try {
      EMError.hasErrorFromResult(result);
    } on EMError catch (e) {
      throw e;
    }
  }

  /// 获取免打扰群组列表
  Future<List<String>?> noDisturbGroupsFromServer() async {
    Map result = await _channel.invokeMethod(ChatMethodKeys.getNoDisturbGroups);
    try {
      EMError.hasErrorFromResult(result);
      _noDisturbGroups =
          result[ChatMethodKeys.getNoDisturbGroups]?.cast<String>();
      return _noDisturbGroups;
    } on EMError catch (e) {
      throw e;
    }
  }
}
